//==========================================================================
//
//      tests/tcp_echo.c
//
//      Simple TCP throughput test - echo component
//
//==========================================================================
//####BSDCOPYRIGHTBEGIN####
//
// -------------------------------------------
//
// Portions of this software may have been derived from OpenBSD or other sources,
// and are covered by the appropriate copyright disclaimers included herein.
//
// -------------------------------------------
//
//####BSDCOPYRIGHTEND####
//==========================================================================
//#####DESCRIPTIONBEGIN####
//
// Author(s):    gthomas
// Contributors: gthomas
// Date:         2000-01-10
// Purpose:      
// Description:  This is the middle part of a three part test.  The idea is
//   to test the throughput of box in a configuration like this:
//
//      +------+   port   +----+     port    +----+
//      |SOURCE|=========>|ECHO|============>|SINK|
//      +------+   9990   +----+     9991    +----+
// 
//
//####DESCRIPTIONEND####
//
//==========================================================================


#include <pkgconf/system.h>
#include <pkgconf/net.h>
#include <cyg/hal/hal_arch.h>



#ifdef CYGBLD_DEVS_ETH_DEVICE_H    // Get the device config if it exists
#include CYGBLD_DEVS_ETH_DEVICE_H  // May provide CYGTST_DEVS_ETH_TEST_NET_REALTIME
#endif

#ifdef CYGPKG_NET_TESTS_USE_RT_TEST_HARNESS // do we use the rt test?
# ifdef CYGTST_DEVS_ETH_TEST_NET_REALTIME // Get the test ancilla if it exists
#  include CYGTST_DEVS_ETH_TEST_NET_REALTIME
# endif
#endif


// Fill in the blanks if necessary
#ifndef TNR_OFF
# define TNR_OFF()
#endif
#ifndef TNR_ON
# define TNR_ON()
#endif
#ifndef TNR_INIT
# define TNR_INIT()
#endif
#ifndef TNR_PRINT_ACTIVITY
# define TNR_PRINT_ACTIVITY()
#endif

#ifndef CYGPKG_KERNEL_SMP_SUPPORT
#define cyg_hal_get_current_cpuid() 0
#define cyg_hal_get_current_threadid() 0
#endif

// Network throughput test code

#include <network.h>

static __inline__ unsigned int
max(unsigned int m, unsigned int n)
{
    return m > n ? m : n;
}

#define SOURCE_PORT1 9990
#define SOURCE_PORT2 9991

#define MAX_BUF 8192
static unsigned char data_buf[MAX_BUF];

struct test_params {
    long nbufs;
    long bufsize;
    long load;
};

struct test_status {
    long ok;
};

#ifndef CYGPKG_LIBC_STDIO
#define perror(s) diag_printf(#s ": %s\n", strerror(errno))
#endif

#define STACK_SIZE (CYGNUM_HAL_STACK_SIZE_TYPICAL + 0x2000)
static char stack[STACK_SIZE*2];
static cyg_thread thread_data1;
static cyg_handle_t thread_handle1;
static cyg_thread thread_data2;
static cyg_handle_t thread_handle2;

// Background load stuff
#define NUM_LOAD_THREADS         20 // Get 5% granularity
#define IDLE_THREAD_PRIORITY     CYGPKG_NET_THREAD_PRIORITY+3
#define LOAD_THREAD_PRIORITY     CYGPKG_NET_THREAD_PRIORITY-3
#define MAIN_THREAD_PRIORITY     CYGPKG_NET_THREAD_PRIORITY-4
#define DESIRED_BACKGROUND_LOAD  50 // should be accurate enough over range

// starting points for load calculation
#define MAX_LOAD_THREAD_LEVEL    100
#define MIN_LOAD_THREAD_LEVEL    0

static cyg_sem_t    init_net_thread_sem;
        

static char         idle_thread_stack[STACK_SIZE];
static cyg_thread   idle_thread_data;
static cyg_handle_t idle_thread_handle;
static cyg_sem_t    idle_thread_sem;
volatile static long long    idle_thread_count;
static cyg_thread   load_thread_data[NUM_LOAD_THREADS];
static cyg_handle_t load_thread_handle[NUM_LOAD_THREADS];
static cyg_sem_t    load_thread_sem[NUM_LOAD_THREADS];
static long         load_thread_level;
static void start_load(int load);
static void do_some_random_computation(int p,int id);
#define abs(n) ((n) < 0 ? -(n) : (n))

static long long no_load_idle_count_1_second;

extern void
cyg_test_exit(void);

void
pexit(char *s)
{
    TNR_OFF();
    perror(s);
    cyg_test_exit();
}

int
do_read(int s, void *_buf, int len)
{
    int total, slen, rlen;
    unsigned char *buf = (unsigned char *)_buf;
    total = 0;
    rlen = len;
    while (total < len) {
        slen = read(s, buf, rlen);
        if (slen != rlen) {
            if (slen < 0) {
                diag_printf("Error after reading %d bytes\n", total);
                return -1;
            }
            rlen -= slen;
            buf += slen;
        }
        total += slen;
    }
    return total;
}

int
do_write(int s, void *_buf, int len)
{
    int total, slen, rlen;
    unsigned char *buf = (unsigned char *)_buf;
    total = 0;
    rlen = len;
    while (total < len) {
        slen = write(s, buf, rlen);
        if (slen != rlen) {
            if (slen < 0) {
                diag_printf("Error after writing %d bytes\n", total);
                return -1;
            }
            rlen -= slen;
            buf += slen;
        }
        total += slen;
    }
    return total;
}



//
// These thread(s) do some amount of "background" computing.  This is used
// to simulate a given load level.  They need to be run at a higher priority 
// than the network code itself.
//
// Like the "idle" thread, they run as long as their "switch" (aka semaphore)
// is enabled.
//
void
net_load(cyg_addrword_t who)
{
    int i;
    while (true) {
        cyg_semaphore_wait(&load_thread_sem[who]);
        for (i = 0;  i < load_thread_level;  i++) {
            do_some_random_computation(i,who);
        }
        cyg_thread_delay(1);  // Wait until the next 'tick'
        cyg_semaphore_post(&load_thread_sem[who]);
    }
}

//
// Some arbitrary computation, designed to use up the CPU and cause associated
// cache "thrash" behaviour - part of background load modelling.
//
static void
do_some_random_computation(int p,int id)
{
    // Just something that might be "hard"
#if 0
    {
        volatile double x;
        x = ((p * 10) * 3.14159) / 180.0;  // radians
    }
#endif
#if 1
    {
        static int footle[0x10001];
        static int counter = 0;
        register int i;

        i = (p << 8) + id + counter++;
        i &= 0xffff;
        footle[ i+1 ] += footle[ i ] + 1;
    }
#endif
}

//
// This thread does nothing but count.  It will be allowed to count
// as long as the semaphore is "free".  
//
void
net_idle(cyg_addrword_t param)
{
    while (true) {
        cyg_semaphore_wait(&idle_thread_sem);
        idle_thread_count++;
        cyg_semaphore_post(&idle_thread_sem);
    }
}

static void
echo_test(cyg_addrword_t p)
{
    int s_source, e_source;
    struct sockaddr_in e_source_addr, e_sink_addr, local;
    int one = 1;
    fd_set in_fds;
    int i, num, len;
    struct test_params params,nparams;
    struct test_status status,nstatus;
    char buf[1024]; int bufpos = 0;
    
    int loopc = 0;
    
    cyg_tick_count_t starttime, stoptime;

    
    s_source = socket(AF_INET, SOCK_STREAM, 0);
    if (s_source < 0) {
        pexit("stream socket");
    }
    if (setsockopt(s_source, SOL_SOCKET, SO_REUSEADDR, &one, sizeof(one))) {
        pexit("setsockopt /source/ SO_REUSEADDR");
    }
    if (setsockopt(s_source, SOL_SOCKET, SO_REUSEPORT, &one, sizeof(one))) {
        pexit("setsockopt /source/ SO_REUSEPORT");
    }
    memset(&local, 0, sizeof(local));
    local.sin_family = AF_INET;
    local.sin_len = sizeof(local);
    local.sin_port = ntohs(p);
    local.sin_addr.s_addr = INADDR_ANY;
    if(bind(s_source, (struct sockaddr *) &local, sizeof(local)) < 0) {
        pexit("bind /source/ error");
    }
    diag_printf("Listem at port %d\n",p);
    listen(s_source, SOMAXCONN);

    while (1) {
      
      e_source = 0;  
      while (true) {
        FD_ZERO(&in_fds);
        FD_SET(s_source, &in_fds);
        num = select(s_source+1, &in_fds, 0, 0, 0);
        if (FD_ISSET(s_source, &in_fds)) {
          len = sizeof(e_source_addr);
          if ((e_source = accept(s_source, (struct sockaddr *)&e_source_addr, &len)) < 0) {
            pexit("accept /source/");
          }
          diag_printf("SOURCE connection from %s:%d\n", 
                      inet_ntoa(e_source_addr.sin_addr), ntohs(e_source_addr.sin_port));
        }
        if ((e_source != 0) ) {
          break;
        }
      }
      
      while(1) {
        char b;
        if ((len = do_read(e_source, &b, 1) != 1)) {
          if (len <= 0) {
            break;
          }
        }
        buf[bufpos++] = b;
        buf[bufpos] = 0;
        if (b == '\n' || bufpos >= 1000){
          unsigned long sp;
          //if (loopc++ > 10) {
            loopc = 0;
            __GET_SP(sp);
            diag_printf("<cpu%d> thread %d:",cyg_hal_get_current_cpuid(),cyg_hal_get_current_threadid());
            //diag_printf("%d\n",cyg_hal_get_current_threadid());
            diag_printf(&buf);
            //}
          
          bufpos = 0;
        }
      }
      
      diag_printf("Connection closed");
    }

    
}

void
net_test(cyg_addrword_t param)
{
    diag_printf("Start TCP test - ECHO mode\n");
    //cyg_thread_delay(1*100);               // Pause for one second
    diag_printf("init Network\n");
    init_all_network_interfaces();
    diag_printf("Network initilized\n");
    cyg_semaphore_post(&init_net_thread_sem);  // Start idle thread
    diag_printf("<cpu%d> thread %d: start",cyg_hal_get_current_cpuid(),cyg_hal_get_current_threadid()); 
    echo_test(SOURCE_PORT1);
    cyg_test_exit();
}

void
net_test2(cyg_addrword_t param)
{
  cyg_semaphore_wait(&init_net_thread_sem);
  diag_printf("<cpu%d> thread %d: start",cyg_hal_get_current_cpuid(),cyg_hal_get_current_threadid()); 
  echo_test(SOURCE_PORT2);
}

void
cyg_start(void)
{
    int i;
    
    cyg_semaphore_init(&init_net_thread_sem, 0);
    // Create a main thread which actually runs the test
    diag_printf("Create thread 1\n");
    cyg_thread_create(MAIN_THREAD_PRIORITY, // Priority
                      net_test,             // entry
                      0,                    // entry parameter
                      "Network test1",       // Name
                      &stack[0],            // Stack
                      STACK_SIZE,           // Size
                      &thread_handle1,       // Handle
                      &thread_data1          // Thread data structure
            );
    cyg_thread_resume(thread_handle1);  // Start it
    diag_printf("stack[0]:0x%x-0x%x\n",&stack[0],
                ((unsigned int)&stack[0])+STACK_SIZE);
    
    // Create the idle thread environment
    diag_printf("Create idle 3\n");
    cyg_thread_create(IDLE_THREAD_PRIORITY,     // Priority
                      net_test2,             // entry
                      0,                        // entry parameter
                      "Network idle",           // Name
                      &idle_thread_stack[0],    // Stack
                      STACK_SIZE,               // Size
                      &idle_thread_handle,      // Handle
                      &idle_thread_data         // Thread data structure
            );
    cyg_thread_resume(idle_thread_handle);      // Start it
    diag_printf("stack[1]:0x%x-0x%x\n",&idle_thread_stack[0],
                ((unsigned int)(&idle_thread_stack[0]))+STACK_SIZE);
   
    cyg_scheduler_start();
}

// EOF tcp_echo.c
